//      LCD

#ifndef _ST7735S_DEFINES
#define _ST7735S_DEFINES

#define ST7735S_NOP                                                                       0x00
#define ST7735S_SWRESET                                                                   0x01
#define ST7735S_RDDID                                                                     0x04
#define ST7735S_RDDST                                                                     0x09
#define ST7735S_RDDPM                                                                     0x0A
#define ST7735S_RDD_MADCTL                                                                0x0B
#define ST7735S_RDD_COLMOD                                                           	  0x0C
#define ST7735S_RDDIM                                                                     0x0D
#define ST7735S_RDDSM                                                                     0x0E

#define ST7735S_SLPIN                                                                     0x10
#define ST7735S_SLPOUT                                                                    0x11
#define ST7735S_PTLON                                                                     0x12
#define ST7735S_NORON                                                                     0x13

#define ST7735S_INVOFF                                                                    0x20
#define ST7735S_INVON                                                                     0x21
#define ST7735S_GAMSET                                                                    0x26
#define ST7735S_DISPOFF                                                                   0x28
#define ST7735S_DISPON                                                                    0x29
#define ST7735S_CASET                                                                     0x2A
#define ST7735S_RASET                                                                     0x2B
#define ST7735S_RAMWR                                                                     0x2C
#define ST7735S_RAMRD                                                                     0x2E

#define ST7735S_PTLAR                                                                     0x30
#define ST7735S_SCRLAR                                                                    0x33
#define ST7735S_TEOFF                                                                     0x34
#define ST7735S_TEON                                                                      0x35
#define ST7735S_MADCTL                                                                    0x36
#define ST7735S_VSCSAD																	  0x37
#define ST7735S_IDMOFF                                                                    0x38
#define ST7735S_IDMON                                                                     0x39
#define ST7735S_COLMOD                                                                    0x3A

#define ST7735S_RDID1                                                                     0xDA
#define ST7735S_RDID2                                                                     0xDB
#define ST7735S_RDID3                                                                     0xDC
#define ST7735S_RDID4                                                                     0xDD

#define ST7735S_FRMCTR1                                                                   0xB1
#define ST7735S_FRMCTR2                                                                   0xB2
#define ST7735S_FRMCTR3                                                                   0xB3
#define ST7735S_INVCTR                                                                    0xB4
#define ST7735S_DISSET5                                                                   0xB6

#define ST7735S_PWCTR1                                                                    0xC0
#define ST7735S_PWCTR2                                                                    0xC1
#define ST7735S_PWCTR3                                                                    0xC2
#define ST7735S_PWCTR4                                                                    0xC3
#define ST7735S_PWCTR5                                                                    0xC4
#define ST7735S_VMCTR1                                                                    0xC5

#define ST7735S_WRID2																	  0xD1
#define ST7735S_WRID3																	  0xD2
#define ST7735S_NVFCTR1																	  0xD9
#define ST7735S_RDID1                                                                     0xDA
#define ST7735S_RDID2                                                                     0xDB
#define ST7735S_RDID3                                                                     0xDC
#define ST7735S_RDID4                                                                     0xDD
#define ST7735S_NVFCTR2																	  0xDE
#define ST7735S_NVFCTR3																	  0xDF

#define ST7735S_PWCTR6                                                                    0xFC

#define ST7735S_GMCTRP1                                                                   0xE0
#define ST7735S_GMCTRN1                                                                   0xE1

#define ST7735S_MADCTL_MY                                                                0x80
#define ST7735S_MADCTL_MX                                                                0x40
#define ST7735S_MADCTL_MV                                                                0x20
#define ST7735S_MADCTL_ML                                                                0x10
#define ST7735S_MADCTL_RGB                                                               0x08
#define ST7735S_MADCTL_MH                                                                0x04

#define ST7735S_TFT_WIDTH                                                                128
#define ST7735S_TFT_LENGTH                                                               160

#define ST7735S_CMD                                                                      0x0
#define ST7735S_DAT                                                                      0x1

enum LCD_COLOR_MODES {
	BIT12PERPIXEL	= 3,
	BIT16PERPIXEL	= 5,
	BIT18PERPIXEL	= 6
};

#define DEFAULT_COLOR_MODE BIT18PERPIXEL

enum LCD_ORIENTATTION_MODES {
	NORMAL					=	0,
	ROTATE_0DG				=	0,
	Y_MIRROR				=	1,
	X_MIRROR				=	2,
	XY_MIROR				=	3,
	ROTATE_180DG			=	3,
	X_Y_EXCHANGE			=	4,
	X_Y_EXCHANGE_Y_MIRROR	=	5,
	ROTATE_270DG			=	5,
	X_Y_EXCHANGE_X_MIRROR	=	6,
	ROTATE_90DG				=	6,
	X_Y_EXCHANGE_XY_MIRROR	=	7
};

#define DEFAULT_ORIENTATTION_MODE ROTATE_0DG

struct LCD_COLORS_STRUCT {
	uint8_t r;
	uint8_t g;
	uint8_t b;
};

typedef struct LCD_COLORS_STRUCT lcdcolors;

union MADCTL_BITS {
	uint8_t ubyte;
	struct {				//--- Low->High bits
		unsigned dummy:2;	//--- Dummy bits
		unsigned vMH:1;		//--- Horizontal refresh order
		unsigned vRGB:1;	//--- Colors bit order
		unsigned vML:1;		//--- Vertical refresh order
		unsigned vMV:1;		//--- Row/Column order
		unsigned vMX:1;		//--- Row address order
		unsigned vMY:1;		//--- Column address order
	} fields;
};

#define IS_CHARACTERS_SUPPORT		  YES

#ifdef IS_CHARACTERS_SUPPORT
#define CHARACTER_WIDTH_SPACE_BEFORE  1
#define CHARACTER_WIDTH				  5
#define CHARACTER_WIDTH_SPACE_AFTER   1
#define CHARACTER_HEIGHT_SPACE_BEFORE 1
#define CHARACTER_HEIGHT			  7
#define CHARACTER_HEIGHT_SPACE_AFTER  1
#define CHARACTER_PLACEHOLDER_WIDTH	  (CHARACTER_WIDTH_SPACE_BEFORE+CHARACTER_WIDTH+CHARACTER_WIDTH_SPACE_AFTER)
#define CHARACTER_PLACEHOLDER_HEIGHT  (CHARACTER_HEIGHT_SPACE_BEFORE+CHARACTER_HEIGHT+CHARACTER_HEIGHT_SPACE_AFTER)

//-- Attention! Characters with CHARACTER_HEIGHT > 16 are not currently supported.
#if CHARACTER_HEIGHT > 16
#error Characters with CHARACTER_HEIGHT > 16 are not supported now.
#endif
#if CHARACTER_HEIGHT > 8
#define CHARACTER_DATA_TYPE			  uint16_t
#define	PGM_READ_DATA				  pgm_read_word
#else 
#define CHARACTER_DATA_TYPE			  uint8_t
#define	PGM_READ_DATA				  pgm_read_byte
#endif
#endif /* IS_CHARACTERS_SUPPORT */

void ST7735S_HWReset(void);
uint8_t ST7735S_DRead(void);
uint8_t ST7735S_Read8(uint8_t cmd);
uint32_t ST7735S_ReadDST(uint8_t cmd);
void ST7735S_Write(const uint8_t fbs, const uint8_t value, const uint8_t cmode, const uint8_t fes);
void SWReset_ST7735S(void);
uint8_t getCurrentColorMode(void);
uint8_t ReadCurrentColorMode(void);
uint8_t getDisplayID1(void);
uint8_t getDisplayID2(void);
void setDisplayID2(const uint8_t id2);
uint8_t getDisplayID3(void);
void setDisplayID3(const uint8_t id3);
uint8_t getDisplayOrders(void);
void setDisplayOrders(const uint8_t orders);
void Init_ST7735S(enum LCD_COLOR_MODES colormode);
void SetWindow(const uint8_t xs, const uint8_t ys, const uint8_t xe, const uint8_t ye);
void WriteColorPixel(const uint8_t rcc, const uint8_t gcc, const uint8_t bcc);
void WriteColorPixels(const uint8_t rcc, const uint8_t gcc, const uint8_t bcc, uint16_t len);
void FillDisplay(const uint8_t rcc, const uint8_t gcc, const uint8_t bcc);
void DrawPixel(const uint8_t x, const uint8_t y, const uint8_t rcc, const uint8_t gcc, const uint8_t bcc);
void DrawHLine(const uint8_t x0, const uint8_t x1, const uint8_t y, const uint8_t rcc, const uint8_t gcc, const uint8_t bcc);
void DrawVLine(const uint8_t x, const uint8_t y0, const uint8_t y1, const uint8_t rcc, const uint8_t gcc, const uint8_t bcc);
void DrawLine(uint8_t x0, uint8_t y0, const uint8_t x1, const uint8_t y1, const uint8_t rcc, const uint8_t gcc, const uint8_t bcc);
void DrawRect(const uint8_t x0, const uint8_t y0, const uint8_t x1, const uint8_t y1, const uint8_t rcc, const uint8_t gcc, const uint8_t bcc);
void DrawFillRect(const uint8_t x0, const uint8_t y0, const uint8_t x1, const uint8_t y1, const uint8_t rcc, const uint8_t gcc, const uint8_t bcc);
void DrawCircleQuadrant(const uint8_t xPos, const uint8_t yPos, const uint8_t radius, const uint8_t quad, const uint8_t rcc, const uint8_t gcc, const uint8_t bcc);
void DrawCircle(const uint8_t xPos, const uint8_t yPos, const uint8_t radius, const uint8_t rcc, const uint8_t gcc, const uint8_t bcc);
void DrawRoundRect(const uint8_t x0, const uint8_t y0, const uint8_t x1, const uint8_t y1, const uint8_t rc, const uint8_t rcc, const uint8_t gcc, const uint8_t bcc);
void DrawFillCircle(const uint8_t xPos, const uint8_t yPos, const uint8_t radius, const uint8_t rcc, const uint8_t gcc, const uint8_t bcc);
void DrawEllipse(const uint8_t x0, const uint8_t y0, const uint8_t width, const uint8_t height, const uint8_t rcc, const uint8_t gcc, const uint8_t bcc);
void DrawFillEllipse(const uint8_t xPos, const uint8_t yPos, const uint8_t width, const uint8_t height, const uint8_t rcc, const uint8_t gcc, const uint8_t bcc);
void SetOrientation(enum LCD_ORIENTATTION_MODES mode);
void definePartArea(const uint8_t psl, const uint8_t pel);
void PartAreaOn(void);
void DefineScrollArea(const uint8_t tfa, const uint8_t vsa, const uint8_t bfa);
void ScrollArea(const uint8_t ssa);
void IdleMode(const uint8_t idlemode);
void setNormalMode(void);
void InversionMode(const uint8_t iversionmode);
void InverseRGB(const uint8_t mode);

#ifdef IS_CHARACTERS_SUPPORT
uint8_t getCurCharX(void);
uint8_t getCurCharY(void);
void GotoXY (const uint8_t, const uint8_t y);
void GotoLine(const uint8_t y);
void WriteChar(const unsigned char ch, const uint8_t x, const uint8_t y, const uint8_t rcc, const uint8_t gcc, const uint8_t bcc, const uint8_t brcc, const uint8_t bgcc, const uint8_t bbcc);
void setNextCharPosition(void);
void WriteStringCB(char *text, const uint8_t rcc, const uint8_t gcc, const uint8_t bcc, const uint8_t brcc, const uint8_t bgcc, const uint8_t bbcc);
void setFGColor(const uint8_t rcc, const uint8_t gcc, const uint8_t bcc);
void setBGColor(const uint8_t brcc, const uint8_t bgcc, const uint8_t bbcc);
void WriteString(char *text);
void WriteStringS(char *text, lcdcolors fgcolor, lcdcolors bgcolor);
void inverseFGColor(void);
void inverseBGColor(void);
void swapFGBGColor(void);
lcdcolors getCurFGColors(void);
lcdcolors getCurBGColors(void);
void setFGSColor(lcdcolors fgcs);
void setBGSColor(lcdcolors bgcs);
#endif /* IS_CHARACTERS_SUPPORT */

#endif /* _ST7735S_DEFINES */
