#include <stdlib.h>
#include <inttypes.h>
#include <util/delay.h>

#include <avr/io.h>
#include <avr/pgmspace.h>
#include <avr/interrupt.h>

#include "ports_config.h"					//---       

#include "ST7735S_DEFINES.H"				//---        LCD

static uint8_t current_color_mode;
static uint8_t current_rotate_mode;
static union MADCTL_BITS MADCTL_States;

#ifdef IS_CHARACTERS_SUPPORT
static uint8_t curX, curY;					//--- Current x,y cursor text position
static lcdcolors curFGColor;				//--- Current foreground text color. Default: WHITE
static lcdcolors curBGColor;				//--- Current background text color. Default: BLACK
#endif

//---    LCD
void ST7735S_HWReset(void) {
    clr_reset;								//--- Pull down reset line
    _delay_us(10);							//--- Wait when IC reconize it
    set_reset;								//--- Pull up reset line
	_delay_ms(120);							//--- Wait when IC internal process finished
}

//---   IC 1 
static void ST7735S_DWrite(uint8_t value) {
	uint8_t i = 8;							//--- 8 ,   
    do {
		if((value & 0x80)) set_sda;			//--- Set data 1 in line
        else               clr_sda;			//--- Set data 0 in line
		asm("nop");							//--- Stabilize data on data line
		set_sck;							//--- Strobe On
		asm("nop");							//--- Stabilize data on strobe line
		clr_sck;							//--- Strobe Off
        value <<= 1;						//--- Shift next data bit
    } while(--i);							//--- Loop do ... while()
	clr_sda;								//---     0?
}

//---   IC 1    /
void ST7735S_Write(const uint8_t fbs, const uint8_t value, const uint8_t cmode, const uint8_t fes) {   
    if (fbs) clr_cs;						//--- Select LCD
    if (cmode) set_rs;						//--- Select Command/Data
    else       clr_rs;
	ST7735S_DWrite(value);
	if (fes) set_cs;						//--- Release LCD
}

uint8_t ST7735S_DRead(void) {
	uint8_t v = 0;							//--- Byte accumulator
	uint8_t i = 8;							//--- 8 bit packet data
	do {
		set_sck;							//--- Strobe On
		v <<= 1;							//--- Shift for next bit and wait for reply
		if (get_sda) v |= 0x01;				//--- Get data bit from sda line
		clr_sck;							//--- Strobe Off
	} while(--i);							//--- Loop do ... while()
	return v;
}

//---  1   IC
 uint8_t ST7735S_Read8(uint8_t cmd) {
	uint8_t v;
	clr_cs;									//--- Select LCD
	clr_rs;									//--- Set command mode transfer
	ST7735S_DWrite(cmd);					//--- Write command to IC
	set_sda_in;								//--- Switch data in recive mode
	clr_sda;								//--- Set in 3-state
	v = ST7735S_DRead();					//--- Read reply from IC
	clr_sda;								//--- Clear data line
	set_sda_out;							//--- Switch data line in transmit mode
	set_cs;									//--- Release LCD
	return v;
}

uint32_t ST7735S_ReadDST(uint8_t cmd) {
	union U32BYTES {
		uint32_t u32;
		struct {
			uint8_t b1;
			uint8_t b2;
			uint8_t b3;
			uint8_t b4;
		} u8;
	} reply;
	clr_cs;									//--- Select LCD
	clr_rs;									//--- Set command mode transfer
	ST7735S_DWrite(ST7735S_RDDST);			//--- Write command to IC
	set_sda_in;								//--- Switch data in recive mode
	clr_sda;								//--- Set in 3-state
	reply.u8.b1 = ST7735S_DRead();			//--- Read reply from IC
	reply.u8.b2 = ST7735S_DRead();			//--- Read reply from IC
	reply.u8.b3 = ST7735S_DRead();			//--- Read reply from IC
	reply.u8.b4 = ST7735S_DRead();			//--- Read reply from IC
	clr_sda;								//--- Clear data line
	set_sda_out;							//--- Switch data line in transmit mode
	set_cs;									//--- Release LCD
	return reply.u32;
}


//---    
uint8_t getCurrentColorMode(void) {
	return current_color_mode;
}

//---   LCD   
uint8_t ReadCurrentColorMode(void) {
	current_color_mode = ST7735S_Read8(ST7735S_RDD_COLMOD);
	return current_color_mode;
}

//---    
void setDisplayOrders(const uint8_t orders) {
	ST7735S_Write(1, ST7735S_MADCTL, ST7735S_CMD, 0);
	ST7735S_Write(0, orders, ST7735S_DAT, 1);
}

//---   LCD   
uint8_t getDisplayOrders(void) {
	MADCTL_States.ubyte = ST7735S_Read8(ST7735S_RDD_MADCTL);
	return MADCTL_States.ubyte;
}
 
// Set the display orientation.
void SetOrientation(enum LCD_ORIENTATTION_MODES mode) {
	if (mode != current_rotate_mode) {
		getDisplayOrders();
		switch (mode) {
			default:
			case NORMAL:
//			case ROTATE_0DG:
				MADCTL_States.fields.vMV = 0;
				MADCTL_States.fields.vMX = 0;
				MADCTL_States.fields.vMY = 0;
				break;
			case X_MIRROR:
				MADCTL_States.fields.vMV = 0;
				MADCTL_States.fields.vMX = 1;
				MADCTL_States.fields.vMY = 0;
				break;
			case Y_MIRROR:
				MADCTL_States.fields.vMV = 0;
				MADCTL_States.fields.vMX = 0;
				MADCTL_States.fields.vMY = 1;
				break;
			case XY_MIROR:
//			case ROTATE_180DG:
				MADCTL_States.fields.vMV = 0;
				MADCTL_States.fields.vMX = 1;
				MADCTL_States.fields.vMY = 1;
				break;
			case X_Y_EXCHANGE:
				MADCTL_States.fields.vMV = 1;
				MADCTL_States.fields.vMX = 0;
				MADCTL_States.fields.vMY = 0;
				break;
			case X_Y_EXCHANGE_Y_MIRROR:
//			case ROTATE_270DG:
				MADCTL_States.fields.vMV = 1;
				MADCTL_States.fields.vMX = 0;
				MADCTL_States.fields.vMY = 1;
				break;
			case X_Y_EXCHANGE_X_MIRROR:
//			case ROTATE_90DG: 
				MADCTL_States.fields.vMV = 1;
				MADCTL_States.fields.vMX = 1;
				MADCTL_States.fields.vMY = 0;
				break;
			case X_Y_EXCHANGE_XY_MIRROR:
				MADCTL_States.fields.vMV = 1;
				MADCTL_States.fields.vMX = 1;
				MADCTL_States.fields.vMY = 1;
				break;
		}
		setDisplayOrders(MADCTL_States.ubyte);
		current_rotate_mode = mode;
	}
}

//---  ID1
uint8_t getDisplayID1(void) {
	return ST7735S_Read8(ST7735S_RDID1);
}

//---  ID2
uint8_t getDisplayID2(void) {
	return (ST7735S_Read8(ST7735S_RDID2) & 0x7F);
}

void setDisplayID2(const uint8_t id2) {
	ST7735S_Write(1, ST7735S_WRID2, ST7735S_CMD, 0);
	ST7735S_Write(0, id2 & 0x7F, ST7735S_DAT, 1);
}

//---  ID3
uint8_t getDisplayID3(void) {
	return ST7735S_Read8(ST7735S_RDID3);
}

void setDisplayID3(const uint8_t id3) {
	ST7735S_Write(1, ST7735S_WRID3, ST7735S_CMD, 0);
	ST7735S_Write(0, id3, ST7735S_DAT, 1);
}

//---   
void SWReset_ST7735S(void) {
	ST7735S_Write(1, ST7735S_SWRESET, ST7735S_CMD, 1);
	_delay_ms(120);
}

//---  LCD    
void Init_ST7735S(enum LCD_COLOR_MODES colormode) {
	ST7735S_HWReset();									// initialize display controller
	asm("nop");
	ST7735S_Write(1, ST7735S_SLPOUT, ST7735S_CMD, 1);	// take display out of sleep mode
	_delay_ms(120); 									// wait 120mS for TFT driver circuits
	ST7735S_Write(1, ST7735S_COLMOD, ST7735S_CMD, 0);	// select color mode
	ST7735S_Write(0, colormode, ST7735S_DAT, 1);
	current_color_mode = colormode;
	ST7735S_Write(1, ST7735S_DISPON, ST7735S_CMD, 1);	// turn display on!
	ReadCurrentColorMode();
	getDisplayOrders();
#ifdef IS_CHARACTERS_SUPPORT
	curFGColor.r = 63; curFGColor.g = 63; curFGColor.b = 63;
	curBGColor.r = 0; curFGColor.g = 0; curFGColor.b = 0;
	curX = 0; curY = 0;
#endif
}

//---     
void SetWindow(const uint8_t xs, const uint8_t ys, const uint8_t xe, const uint8_t ye) {
    ST7735S_Write(1, ST7735S_CASET, ST7735S_CMD, 0);
    ST7735S_Write(0, 00, ST7735S_DAT, 0);
    ST7735S_Write(0, xs, ST7735S_DAT, 0);
    ST7735S_Write(0, 00, ST7735S_DAT, 0);
    ST7735S_Write(0, xe, ST7735S_DAT, 1);
    ST7735S_Write(1, ST7735S_RASET, ST7735S_CMD, 0);
    ST7735S_Write(0, 00, ST7735S_DAT, 0);
    ST7735S_Write(0, ys, ST7735S_DAT, 0);
    ST7735S_Write(0, 00, ST7735S_DAT, 0);
    ST7735S_Write(0, ye, ST7735S_DAT, 1);
}

//---   1  
void WriteColorPixel(const uint8_t rcc, const uint8_t gcc, const uint8_t bcc) {
	uint8_t cb1, cb2, cb3;
	switch (current_color_mode) {
		case BIT18PERPIXEL:
			cb1 = (0B00111111 & rcc)<<2;
			cb2 = (0B00111111 & gcc)<<2;
			cb3	= (0B00111111 & bcc)<<2;
			break;
		case BIT16PERPIXEL:
			cb1 = (0B00011111 & rcc)<<3;
			cb3 = (0B00111111 & gcc);
			cb1 = cb1 | (cb3>>3);
			cb2 = (cb3<<5) | (0B00011111 & bcc);
			break;
		case BIT12PERPIXEL:
			cb1 = ((0B00001111 & rcc)<<4) | (0B00001111 & gcc);
			cb2 = ((0B00001111 & bcc)<<4) | (0B00001111 & rcc);
			cb3 = ((0B00001111 & gcc)<<4) | (0B00001111 & bcc);
			break;
		default:
			cb1 = 0; cb2 = 0; cb3 = 0;
			break;
	}
	ST7735S_DWrite(cb1);
	ST7735S_DWrite(cb2);
	if (current_color_mode != BIT16PERPIXEL) ST7735S_DWrite(cb3);
}

//---  'len'   
void WriteColorPixels(const uint8_t rcc, const uint8_t gcc, const uint8_t bcc, uint16_t len) {
	uint8_t cb1, cb2, cb3;
	switch (current_color_mode) {
		case BIT18PERPIXEL:
			cb1 = (0B00111111 & rcc)<<2;
			cb2 = (0B00111111 & gcc)<<2;
			cb3	= (0B00111111 & bcc)<<2;
			break;
		case BIT16PERPIXEL:
			cb1 = (0B00011111 & rcc)<<3;
			cb3 = (0B00111111 & gcc);
			cb1 = cb1 | (cb3>>3);
			cb2 = (cb3<<5) | (0B00011111 & bcc);
			break;
		case BIT12PERPIXEL:
			cb1 = ((0B00001111 & rcc)<<4) | (0B00001111 & gcc);
			cb2 = ((0B00001111 & bcc)<<4) | (0B00001111 & rcc);
			cb3 = ((0B00001111 & gcc)<<4) | (0B00001111 & bcc);
			break;
		default:
			cb1 = 0; cb2 = 0; cb3 = 0;
			break;
	}
	ST7735S_Write(1, ST7735S_RAMWR, ST7735S_CMD, 1);		//--- Begin write data to memory
	clr_cs;													//--- Select LCD
	set_rs;													//--- Data transfer only
	do {
		ST7735S_DWrite(cb1);
		ST7735S_DWrite(cb2);
		if (current_color_mode != BIT16PERPIXEL) { 
			ST7735S_DWrite(cb3);
			if (current_color_mode == BIT12PERPIXEL) --len;
		}
	} while(--len);
	clr_rs;													//--- Back to command mode
	set_cs;													//--- Release LCD.
}

//---     
void FillDisplay(const uint8_t rcc, const uint8_t gcc, const uint8_t bcc) {
	SetWindow(0, 0,
			  (MADCTL_States.fields.vMV ? (ST7735S_TFT_LENGTH - 1) : (ST7735S_TFT_WIDTH - 1)),
			  (MADCTL_States.fields.vMV ? (ST7735S_TFT_WIDTH - 1) : (ST7735S_TFT_LENGTH - 1)));
	WriteColorPixels(rcc, gcc, bcc, ST7735S_TFT_LENGTH * ST7735S_TFT_WIDTH);
}

//---     
void DrawPixel(const uint8_t x, const uint8_t y, const uint8_t rcc, const uint8_t gcc, const uint8_t bcc) {
	SetWindow(x, y, x, y);
	ST7735S_Write(1, ST7735S_RAMWR, ST7735S_CMD, 1);				//--- Begin write data to memory
	clr_cs;								//--- Select LCD
	set_rs;								//--- Data transfer only
	WriteColorPixel(rcc, gcc, bcc);
	clr_rs;								//--- Back to command mode
	set_cs;								//--- Release LCD.
}

//---   
void DrawHLine(const uint8_t x0, const uint8_t x1, const uint8_t y,
			   const uint8_t rcc, const uint8_t gcc, const uint8_t bcc) {	// draws a horizontal line in given color
	SetWindow(x0, y, x1, y);
	WriteColorPixels(rcc, gcc, bcc, x1 - x0 + 1);
}

//---   
void DrawVLine(const uint8_t x, const uint8_t y0, const uint8_t y1,
			   const uint8_t rcc, const uint8_t gcc, const uint8_t bcc) {	// draws a vertical line in given color
	SetWindow(x, y0, x, y1);
	WriteColorPixels(rcc, gcc, bcc, y1 - y0 + 1);
}

//---   
void DrawLine(uint8_t x0, uint8_t y0, const uint8_t x1, const uint8_t y1,
			  const uint8_t rcc, const uint8_t gcc, const uint8_t bcc) {
	int dx = abs(x1-x0), sx = x0<x1 ? 1 : -1;
	int dy = abs(y1-y0), sy = y0<y1 ? 1 : -1;
	int error = (dx>dy ? dx : -dy)/2, e2;
	while (1) {
		DrawPixel(x0, y0, rcc, gcc, bcc);
		if (x0==x1 && y0==y1) break;
		e2 = error;
		if (e2 >-dx) { error -= dy; x0 += sx; }
		if (e2 < dy) { error += dx; y0 += sy; }
	}
}

//---    
void DrawRect(const uint8_t x0, const uint8_t y0, const uint8_t x1, const uint8_t y1,
			  const uint8_t rcc, const uint8_t gcc, const uint8_t bcc) {	// draws a rectangle in given color
	DrawHLine(x0, x1, y0, rcc, gcc, bcc);
	DrawHLine(x0, x1, y1, rcc, gcc, bcc);
	DrawVLine(x0, y0, y1, rcc, gcc, bcc);
	DrawVLine(x1, y0, y1, rcc, gcc, bcc);
}

//---     
void DrawFillRect(const uint8_t x0, const uint8_t y0, const uint8_t x1, const uint8_t y1,
			      const uint8_t rcc, const uint8_t gcc, const uint8_t bcc) {
	SetWindow(x0, y0, x1, y1);
	WriteColorPixels(rcc, gcc, bcc, (x1-x0+1)*(y1-y0+1));
}

static unsigned long intsqrt(const unsigned long val) {	// calculate integer value of square root
	unsigned long mulMask = 0x0008000;
	unsigned long retVal = 0;
	if (val > 0) {
		while(mulMask != 0) {
			retVal |= mulMask;
			if ((retVal * retVal) > val) retVal &= ~ mulMask;
			mulMask >>= 1;
		}
	}
	return retVal;
}

static uint16_t roundintsqrt(const uint16_t val) {
	unsigned long vsqrt = intsqrt(val << 2); 	// sqrt(v*4) = sqrt(v)*2; v*4 = v << 2
	uint16_t v = vsqrt >> 1;
	if (vsqrt & 1UL) ++v;
	return v;
}

// draws circle quadrant(s) centered at x,y with given radius & color
// quad is a bit-encoded representation of which cartesian quadrant(s) to draw.
// Remember that the y axis on our display is 'upside down':
// bit 0: draw quadrant I (lower right)
// bit 1: draw quadrant IV (upper right)
// bit 2: draw quadrant II (lower left)
// bit 3: draw quadrant III (upper left)
void DrawCircleQuadrant(const uint8_t xPos, const uint8_t yPos,
						const uint8_t radius, const uint8_t quad,
						const uint8_t rcc, const uint8_t gcc, const uint8_t bcc) {
	uint8_t x, xEnd = (uint8_t)((uint16_t)((707*radius))/1000 + 1);
	for (x=0; x < xEnd; x++) {
		uint8_t y = (uint8_t)roundintsqrt(radius*radius - x*x);
		if (quad & 0x01) {
			DrawPixel(xPos+x, yPos+y, rcc, gcc, bcc); // lower right
			DrawPixel(xPos+y, yPos+x, rcc, gcc, bcc);
		}
		if (quad & 0x02) {
			DrawPixel(xPos+x, yPos-y, rcc, gcc, bcc); // upper right
			DrawPixel(xPos+y, yPos-x, rcc, gcc, bcc);
		}
		if (quad & 0x04) {
			DrawPixel(xPos-x, yPos+y, rcc, gcc, bcc); // lower left
			DrawPixel(xPos-y, yPos+x, rcc, gcc, bcc);
		}
		if (quad & 0x08) {
			DrawPixel(xPos-x, yPos-y, rcc, gcc, bcc); // upper left
			DrawPixel(xPos-y, yPos-x, rcc, gcc, bcc);
		}
	}
}

// draws circle at x,y with given radius & color
void DrawCircle(const uint8_t xPos, const uint8_t yPos, const uint8_t radius,
				const uint8_t rcc, const uint8_t gcc, const uint8_t bcc) {
	DrawCircleQuadrant(xPos, yPos, radius, 0x0F, rcc, gcc, bcc); // do all 4 quadrants
}

// draws a rounded rectangle with corner radius r.
// coordinates: top left = x0,y0; bottom right = x1,y1
void DrawRoundRect(const uint8_t x0, const uint8_t y0, const uint8_t x1, const uint8_t y1,
				   const uint8_t rc, const uint8_t rcc, const uint8_t gcc, const uint8_t bcc) {
	DrawHLine(x0+rc, x1-rc, y0, rcc, gcc, bcc); // top side
	DrawHLine(x0+rc, x1-rc, y1, rcc, gcc, bcc); // bottom side
	DrawVLine(x0, y0+rc, y1-rc, rcc, gcc, bcc); // left side
	DrawVLine(x1, y0+rc, y1-rc, rcc, gcc, bcc); // right side
	DrawCircleQuadrant(x0+rc, y0+rc, rc, 8, rcc, gcc, bcc); // upper left corner
	DrawCircleQuadrant(x1-rc, y0+rc, rc, 2, rcc, gcc, bcc); // upper right corner
	DrawCircleQuadrant(x0+rc, y1-rc, rc, 4, rcc, gcc, bcc); // lower left corner
	DrawCircleQuadrant(x1-rc, y1-rc, rc, 1, rcc, gcc, bcc); // lower right corner
}

// draws filled circle at x,y with given radius & color
void DrawFillCircle(const uint8_t xPos, const uint8_t yPos, const uint8_t radius,
				    const uint8_t rcc, const uint8_t gcc, const uint8_t bcc) {
	uint16_t x, r2 = radius * radius;
	for (x=0; x<=radius; x++) {
		uint8_t y = roundintsqrt(r2-x*x);
		uint8_t y0 = yPos-y;
		uint8_t y1 = yPos+y;
		DrawVLine(xPos+x, y0, y1, rcc, gcc, bcc);
		DrawVLine(xPos-x, y0, y1, rcc, gcc, bcc);
	}
}

// draws an ellipse of given width & height two-part Bresenham method
// note: slight discontinuity between parts on some (narrow) ellipses.
void DrawEllipse(const uint8_t x0, const uint8_t y0, const uint8_t width, const uint8_t height,
				 const uint8_t rcc, const uint8_t gcc, const uint8_t bcc) {
	uint8_t a=width/2, b=height/2;
	uint8_t x = 0, y = b;
	long a2 = (long)a*a*2;
	long b2 = (long)b*b*2;
	long error = (long)a*a*b;
	long stopY=0, stopX = a2*b;
	while (stopY <= stopX) {
		DrawPixel(x0+x, y0+y, rcc, gcc, bcc);
		DrawPixel(x0+x, y0-y, rcc, gcc, bcc);
		DrawPixel(x0-x, y0+y, rcc, gcc, bcc);
		DrawPixel(x0-x, y0-y, rcc, gcc, bcc);
		++x;
		error -= b2*(x-1);
		stopY += b2;
		if (error < 0) {
			error += a2*(y-1);
			--y;
			stopX -= a2;
		}
	}
	x=a; y=0; error = b*b*a;
	stopY = a*b2; stopX = 0;
	while (stopY >= stopX) {
		DrawPixel(x0+x, y0+y, rcc, gcc, bcc);
		DrawPixel(x0+x, y0-y, rcc, gcc, bcc);
		DrawPixel(x0-x, y0+y, rcc, gcc, bcc);
		DrawPixel(x0-x, y0-y, rcc, gcc, bcc);
		++y;
		error -= a2*(y-1);
		stopX += a2;
		if (error < 0) {
			error += b2*(x-1);
			--x;
			stopY -= b2;
		}
	}
}

// draws a filled ellipse of given width & height
void DrawFillEllipse(const uint8_t xPos, const uint8_t yPos, const uint8_t width, const uint8_t height, const uint8_t rcc, const uint8_t gcc, const uint8_t bcc) {
	uint8_t a=width/2, b=height/2; // get x & y radii
	uint8_t x1, x0 = a, y = 1, dx = 0;
	long a2 = a*a, b2 = b*b; // need longs: big numbers!
	long a2b2 = a2 * b2;
	DrawHLine(xPos-a, xPos+a, yPos, rcc, gcc, bcc); // draw centerline
	while (y<=b) { // draw horizontal lines...
		for (x1= x0-(dx-1); x1>0; x1--) if (b2*x1*x1 + a2*y*y <= a2b2) break;
		dx = x0-x1;
		x0 = x1;
		DrawHLine(xPos-x0, xPos+x0, yPos+y, rcc, gcc, bcc);
		DrawHLine(xPos-x0, xPos+x0, yPos-y, rcc, gcc, bcc);
		++y;
	}
}

void definePartArea(const uint8_t psl, const uint8_t pel) {
	ST7735S_Write(1, ST7735S_PTLAR, ST7735S_CMD, 0);
	ST7735S_Write(0, 0, ST7735S_DAT, 0);
    ST7735S_Write(0, psl, ST7735S_DAT, 0);
    ST7735S_Write(0, 0, ST7735S_DAT, 0);
    ST7735S_Write(0, pel, ST7735S_DAT, 1);
}

void PartAreaOn(void) {
	ST7735S_Write(1, ST7735S_PTLON, ST7735S_CMD, 1);
}

void DefineScrollArea(const uint8_t tfa, const uint8_t vsa, const uint8_t bfa) {
	ST7735S_Write(1, ST7735S_SCRLAR, ST7735S_CMD, 0);
	ST7735S_Write(0, 0, ST7735S_DAT, 0);
    ST7735S_Write(0, tfa, ST7735S_DAT, 0);
    ST7735S_Write(0, 0, ST7735S_DAT, 0);
    ST7735S_Write(0, vsa, ST7735S_DAT, 0);
	ST7735S_Write(0, 0, ST7735S_DAT, 0);
    ST7735S_Write(0, bfa, ST7735S_DAT, 1);
}

void ScrollArea(const uint8_t ssa) {
	ST7735S_Write(1, ST7735S_VSCSAD, ST7735S_CMD, 0);
	ST7735S_Write(0, ssa, ST7735S_DAT, 1);
}

void IdleMode(const uint8_t idlemode) {
	ST7735S_Write(1, idlemode ? ST7735S_IDMON : ST7735S_IDMOFF, ST7735S_CMD, 1);
}

void setNormalMode(void) {
	ST7735S_Write(1, ST7735S_NORON, ST7735S_CMD, 1);
}

void InversionMode(const uint8_t iversionmode) {
	ST7735S_Write(1, iversionmode ? ST7735S_INVON : ST7735S_INVOFF, ST7735S_CMD, 1);
}

void InverseRGB(const uint8_t mode) {
	MADCTL_States.fields.vRGB = mode ? 1 : 0;
	setDisplayOrders(MADCTL_States.ubyte);
}

//      .
//   ,   .

#ifdef IS_CHARACTERS_SUPPORT

// --------------------------------------
// Text out functions and data structures.
// --------------------------------------

#include "symbol_characters.h"

// write char to display X,Y coordinates using ASCII CHARACTER_WIDTH x CHARACTER_HEIGHT font in NVRAM
void WriteChar(const unsigned char ch, const uint8_t x, const uint8_t y,
			   const uint8_t rcc, const uint8_t gcc, const uint8_t bcc,		//--- draw color
			   const uint8_t brcc, const uint8_t bgcc, const uint8_t bbcc) {//--- back color
	if ((ch >= START_LOW_ASCII_SYMBOLS && ch <= END_LOW_ASCII_SYMBOLS) 
#ifdef HIGH_ASCII_SYMBOLS
		|| (ch >= START_HIGH_ASCII_SYMBOLS && ch <= END_HIGH_ASCII_SYMBOLS)
#endif
		) {
		uint8_t row, col;										//---   
		CHARACTER_DATA_TYPE symbol[ CHARACTER_WIDTH ];			//---   
		CHARACTER_DATA_TYPE mask = 1;							//---     
		for (col=0; col < CHARACTER_WIDTH; col++)				//---     
#ifdef HIGH_ASCII_SYMBOLS
			symbol[col] = PGM_READ_DATA(&(FONT_CHARS
				[(
					ch >= START_HIGH_ASCII_SYMBOLS ? 
					(ch-START_HIGH_ASCII_SYMBOLS+END_LOW_ASCII_SYMBOLS+1-START_LOW_ASCII_SYMBOLS) : 
					(ch-START_LOW_ASCII_SYMBOLS)
				)][col]));
#else
			symbol[col] = PGM_READ_DATA(&(FONT_CHARS[ch-START_LOW_ASCII_SYMBOLS][col]));
#endif
		SetWindow(x, y, x + CHARACTER_PLACEHOLDER_WIDTH - 1, y + CHARACTER_PLACEHOLDER_HEIGHT - 1);
		ST7735S_Write(1, ST7735S_RAMWR, ST7735S_CMD, 1);		//--- Begin write data to memory
		clr_cs;													//--- Select LCD
		set_rs;

#if CHARACTER_HEIGHT_SPACE_BEFORE > 0	
		for(row=0; row < CHARACTER_HEIGHT_SPACE_BEFORE*CHARACTER_PLACEHOLDER_WIDTH; ++row) {
			WriteColorPixel(brcc, bgcc, bbcc);					//--- Back color
		}
#endif

		for (row=0; row < CHARACTER_HEIGHT; ++row) {			//---     
#if CHARACTER_WIDTH_SPACE_BEFORE > 0
			for (col=0; col < CHARACTER_WIDTH_SPACE_BEFORE; ++col) {	//--- Draw right free space	
				WriteColorPixel(brcc, bgcc, bbcc);				//--- Back color
			}
#endif
			for (col=0; col < CHARACTER_WIDTH; ++col) {			//---     
				if (symbol[col] & mask) WriteColorPixel(rcc, gcc, bcc);	
				else					WriteColorPixel(brcc, bgcc, bbcc);
			}
#if CHARACTER_WIDTH_SPACE_AFTER > 0
			for (col=0; col < CHARACTER_WIDTH_SPACE_AFTER; ++col) {	//--- Draw left free space	
				WriteColorPixel(brcc, bgcc, bbcc);				//--- Back color
			}
#endif
			mask <<= 1;
		}
#if CHARACTER_HEIGHT_SPACE_AFTER > 0
		for(row=0; row < CHARACTER_HEIGHT_SPACE_AFTER*CHARACTER_PLACEHOLDER_WIDTH; ++row) {
			WriteColorPixel(brcc, bgcc, bbcc);					//--- Back color
		}
#endif
		clr_rs;													//--- Back to command mode
		set_cs;													//--- Release LCD.
	}
}

uint8_t getCurCharX(void) { return curX; }
uint8_t getCurCharY(void) { return curY; }
 
uint8_t getMaxCharX(void) { 
	return(MADCTL_States.fields.vMV ? 
		(ST7735S_TFT_LENGTH/CHARACTER_PLACEHOLDER_WIDTH - 1)
		:
		(ST7735S_TFT_WIDTH/CHARACTER_PLACEHOLDER_WIDTH - 1)
	);
}

uint8_t getMaxCharY(void) {
	return(MADCTL_States.fields.vMV ? 
		(ST7735S_TFT_WIDTH / CHARACTER_PLACEHOLDER_HEIGHT - 1)
		:
		(ST7735S_TFT_LENGTH / CHARACTER_PLACEHOLDER_HEIGHT - 1)
	);
}

void setNextCharPosition(void) {
	if (++curX > getMaxCharX()) {
		curX = 0;
		if (++curY > getMaxCharY()) curY = 0;
	}
}

// position cursor on character x,y grid
void GotoXY (const uint8_t x, const uint8_t y) {
	if (x <= getMaxCharX() && y <= getMaxCharY()) {
		curX = x; curY = y;
	}
}

// position character cursor to start of line y
void GotoLine(const uint8_t y) {
	if (y <= getMaxCharY()) {
		curX = 0; curY = y;
	}
}

lcdcolors getCurFGColors(void) { return curFGColor; }
lcdcolors getCurBGColors(void) { return curBGColor; }

void setFGColor(const uint8_t rcc, const uint8_t gcc, const uint8_t bcc) {
	curFGColor.r = rcc;
	curFGColor.g = gcc;
	curFGColor.b = bcc;
}

void setFGSColor(lcdcolors fgcs) {
//	curFGColor.r = fgcs.r;
//	curFGColor.g = fgcs.g;
//	curFGColor.b = fgcs.b;
	curFGColor = fgcs;
}

void setBGSColor(lcdcolors bgcs) {
//	curBGColor.r = bgcs.r;
//	curBGColor.g = bgcs.g;
//	curBGColor.b = bgcs.b;
	curBGColor = bgcs;
}

void setBGColor(const uint8_t brcc, const uint8_t bgcc, const uint8_t bbcc) {
	curBGColor.r = brcc;
	curBGColor.g = bgcc;
	curBGColor.b = bbcc;
}

void inverseFGColor(void) {
	curFGColor.r = ~curFGColor.r;
	curFGColor.g = ~curFGColor.g;
	curFGColor.b = ~curFGColor.b;
}

void inverseBGColor(void) {
	curBGColor.r = ~curBGColor.r;
	curBGColor.g = ~curBGColor.g;
	curBGColor.b = ~curBGColor.b;
}

void swapFGBGColor(void) {
	uint8_t rc = curFGColor.r;
	uint8_t gc = curFGColor.g;
	uint8_t bc = curFGColor.b;
	curFGColor.r = curBGColor.r;
	curFGColor.g = curBGColor.g;
	curFGColor.b = curBGColor.b;
	curBGColor.r = rc;
	curBGColor.g = gc;
	curBGColor.b = bc;
}

void WriteStringCB(char *text,
				   const uint8_t rcc, const uint8_t gcc, const uint8_t bcc,			//--- draw color
				   const uint8_t brcc, const uint8_t bgcc, const uint8_t bbcc) {	//--- back color
	for (; *text; ++text) {
		WriteChar(*text,
				  curX * CHARACTER_PLACEHOLDER_WIDTH,
				  curY * CHARACTER_PLACEHOLDER_HEIGHT,
				  rcc, gcc, bcc, brcc, bgcc, bbcc);	// write the char
		setNextCharPosition();
	}
}

void WriteStringS(char *text, lcdcolors fgcolor, lcdcolors bgcolor) {
	for (; *text; ++text) {
		WriteChar(*text,
				  curX * CHARACTER_PLACEHOLDER_WIDTH,
				  curY * CHARACTER_PLACEHOLDER_HEIGHT,
				  fgcolor.r, fgcolor.g, fgcolor.b, bgcolor.r, bgcolor.g, bgcolor.b);	// write the char
		setNextCharPosition();
	}
}

void WriteString(char *text) {
	for (; *text; ++text) {
		WriteChar(*text,
				  curX * CHARACTER_PLACEHOLDER_WIDTH,
				  curY * CHARACTER_PLACEHOLDER_HEIGHT,
				  curFGColor.r, curFGColor.g, curFGColor.b,
				  curBGColor.r, curBGColor.g, curBGColor.b);	// write the char
		setNextCharPosition();
	}
}

#endif	/* IS_CHARACTERS_SUPPORT */